/* adxl362.c - ADXL362 Three-Axis Digital Accelerometers */

/*
 * Copyright (c) 2017 IpTronix S.r.l.
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#define DT_DRV_COMPAT vesper_inmp621

#include <kernel.h>
#include <string.h>
#include <init.h>
#include <drivers/gpio.h>
#include <sys/byteorder.h>
#include <sys/__assert.h>
#include <drivers/i2c.h>
#include <audio/dmic.h>
#include "nrfx_pdm.h"

#include "inmp621.h"

#include <logging/log.h>
LOG_MODULE_REGISTER(inmp621, LOG_LEVEL_DBG);

#define PDM_PERIPHERAL_RATIO_64 64UL
#define PDM_PERIPHERAL_RATIO_80 80UL

static struct inmp621_data inmp621_data;
static int16_t pcm_buffer[2][CONFIG_INMP621_PDM_BUFFER_SIZE];
static uint8_t pcm_buffer_in_use = 0;
static bool pcm_buffer_avalable = false;

static const struct inmp621_config inmp621_cfg = {
	
	.data_pin = DT_INST_PROP(0, data_pin),
	.clk_pin = DT_INST_PROP(0, clk_pin),
	.lr_pin_level = DT_INST_PROP(0, lr_select),
#if defined(CONFIG_INMP621_INT)
	.gpio_port = DT_INST_GPIO_LABEL(0, dout_gpios),
	.dout_pin = DT_INST_GPIO_PIN(0, dout_gpios),
	.dout_flags = DT_INST_GPIO_FLAGS(0, dout_gpios),
#endif
};


static int inmp621_wdt_configure(const struct device *dev, uint8_t wdt_dly, bool enable)
{
	uint8_t value = (wdt_dly << INMP621_WDT_DELAY_Pos) & INMP621_WDT_DELAY_Msk;
	uint8_t buff;
	int ret;

	if(enable){
		value |= 0xFF & INMP621_WDT_ENABLE_Msk;
	}else{
		value |= 0x0 & INMP621_WDT_ENABLE_Msk;
	}


	buff &= ~(INMP621_WDT_DELAY_Msk | INMP621_WDT_ENABLE_Msk);
	buff |= value;

	return 0;
}

static int inmp621_wos_pga_min_thr_set(const struct device *dev, float pga_gain_db)
{
	uint8_t pga_min_gain;
	uint8_t buff;
	int ret;

	ret = inmp621_get_reg(dev, INMP621_REG_WOS_PGA_MIN_THR, &buff);
	if (ret < 0) {
		LOG_ERR("Couldn't read register %d, error %d", INMP621_REG_WOS_PGA_MIN_THR, ret);
		return ret;
	}

	pga_min_gain = INMP621_DB_TO_PGA(pga_gain_db);

	buff &= ~INMP621_WOS_PGA_MIN_THR_Msk;
	buff |= pga_min_gain & INMP621_WOS_PGA_MIN_THR_Msk;

	ret = inmp621_set_reg(dev, INMP621_REG_WOS_PGA_MIN_THR, buff);
	if (ret < 0) {
		LOG_ERR("Couldn't write register %d, error %d", INMP621_REG_WOS_PGA_MIN_THR, ret);
		return ret;
	}

	return 0;
}

static int inmp621_wos_pga_max_thr_set(const struct device *dev, float pga_gain_db)
{
	uint8_t pga_max_gain;
	uint8_t buff;
	int ret;

	ret = inmp621_get_reg(dev, INMP621_REG_WOS_PGA_MAX_THR, &buff);
	if (ret < 0) {
		LOG_ERR("Couldn't read register %d, error %d", INMP621_REG_WOS_PGA_MAX_THR, ret);
		return ret;
	}

	pga_max_gain = INMP621_DB_TO_PGA(pga_gain_db);
	
	buff &= ~INMP621_WOS_PGA_MAX_THR_Msk;
	buff |= pga_max_gain & INMP621_WOS_PGA_MAX_THR_Msk;

	ret = inmp621_set_reg(dev, INMP621_REG_WOS_PGA_MAX_THR, buff);
	if (ret < 0) {
		LOG_ERR("Couldn't write register %d, error %d", INMP621_REG_WOS_PGA_MAX_THR, ret);
		return ret;
	}

	return 0;
}

static int inmp621_bpf_set( const struct device *dev, uint8_t wos_lpf_freq, uint8_t wos_hpf_freq)
{
	uint8_t buff;
	int ret;

	buff = (wos_lpf_freq & INMP621_WOS_BANDPASS_LPF_Msk) |
		   ((wos_hpf_freq << INMP621_WOS_BANDPASS_HPF_Pos) & INMP621_WOS_BANDPASS_HPF_Msk);

	ret = inmp621_set_reg(dev, INMP621_REG_WOS_FILTER, buff);
	if (ret < 0) {
		LOG_ERR("Couldn't write register %d, error %d", INMP621_REG_WOS_FILTER, ret);
		return ret;
	}

	return 0;

}

static int inmp621_fast_mode_cnt_set( const struct device *dev, uint8_t fast_mode_cnt)
{
	uint8_t buff;
	int ret;

	ret = inmp621_get_reg(dev, INMP621_REG_WOS_PGA_MIN_THR, &buff);
	if (ret < 0) {
		LOG_ERR("Couldn't read register %d, error %d", INMP621_REG_WOS_PGA_MIN_THR, ret);
		return ret;
	}

	buff &= ~INMP621_FAST_MODE_CNT_Msk;
	buff |= (fast_mode_cnt << INMP621_FAST_MODE_CNT_Pos) & INMP621_FAST_MODE_CNT_Msk;

	ret = inmp621_set_reg(dev, INMP621_REG_WOS_PGA_MIN_THR, buff);
	if (ret < 0) {
		LOG_ERR("Couldn't write register %d, error %d", INMP621_REG_WOS_PGA_MIN_THR, ret);
		return ret;
	}

	return 0;
}

static int inmp621_wos_rms_set( const struct device *dev, uint8_t value)
{
	uint8_t buff;
	int ret;

	ret = inmp621_get_reg(dev, INMP621_REG_WOS_PGA_MAX_THR, &buff);
	if (ret < 0) {
		LOG_ERR("Couldn't read register %d, error %d", INMP621_REG_WOS_PGA_MAX_THR, ret);
		return ret;
	}

	buff &= ~INMP621_WOS_RMS_Msk;
	buff |= (value << INMP621_WOS_RMS_Pos) & INMP621_WOS_RMS_Msk;

	ret = inmp621_set_reg(dev, INMP621_REG_WOS_PGA_MAX_THR, buff);
	if (ret < 0) {
		LOG_ERR("Couldn't write register %d, error %d", INMP621_REG_WOS_PGA_MAX_THR, ret);
		return ret;
	}

	return 0;
}

static int inmp621_wos_dout_thresh_set( const struct device *dev, uint8_t wos_thresh)
{
	uint8_t buff;
	int ret;

	/* WOS_THRESH register value should not be 0 */
	if(!wos_thresh)
	{
		LOG_WRN("Invalid WOS_THRESH value");
		return -EINVAL;
	}

	buff = wos_thresh & INMP621_WOS_THRESHOLD_Msk;

	ret = inmp621_set_reg(dev, INMP621_REG_WOS_THRESHOLD, buff);
	if (ret < 0) {
		LOG_ERR("Couldn't write register %d, error %d", INMP621_REG_WOS_THRESHOLD, ret);
		return ret;
	}

	return 0;
}

static int inmp621_amb_sound_lvl_get(const struct device *dev, float *amb_sound_lvl)
{
	uint8_t buff;
	int ret;

	ret = inmp621_get_reg(dev, INMP621_REG_WOS_PGA_GAIN, &buff);
	if (ret < 0) {
		LOG_ERR("Couldn't read register %d, error %d", INMP621_REG_WOS_PGA_GAIN, ret);
		return ret;
	}

	/* Translate PGA gain to DB using the LUT */
	*amb_sound_lvl = INMP621_PGA_TO_DB(buff);


	return 0;
}

static void inmp621_pdm_handler(nrfx_pdm_evt_t const * p_evt)
{
	uint32_t ret;
	if(p_evt->error != NRFX_PDM_NO_ERROR)
	{
		LOG_ERR("PDM buffer overflow");
	}

	if(p_evt->buffer_requested){
		pcm_buffer_in_use ^= 1;
		ret = nrfx_pdm_buffer_set(&pcm_buffer[pcm_buffer_in_use][0], CONFIG_INMP621_PDM_BUFFER_SIZE);
		if(ret != NRFX_SUCCESS)
		{
			LOG_ERR("Failed to set buffer, error %d", ret);
		}
	}
	if(p_evt->buffer_released){
		pcm_buffer_avalable = true;
		//do something with released buffer
	}
}

static int inmp621_configure(const struct device *dev,
				 struct dmic_cfg *cfg)
{
	int ret;
	struct inmp621_data *dmic_data = dev->data;
	const struct inmp621_config *config = dev->config;
	uint32_t audio_sample_rate;

	/* Initialise with default parameters */
	nrfx_pdm_config_t pdm_param = 
			NRFX_PDM_DEFAULT_CONFIG(config->clk_pin, config->data_pin);

	/* Configure NRFX PDM struct parameters */
	pdm_param.mode = INMP621_PDM_MODE;
	if(pdm_param.mode == NRF_PDM_MODE_STEREO){
		/* If stereo, left channel always on falling CLK edge */
		pdm_param.gain_l = NRF_PDM_GAIN_DEFAULT;
		pdm_param.gain_r = NRF_PDM_GAIN_DEFAULT;
		pdm_param.edge = NRF_PDM_EDGE_LEFTFALLING;
	}else{
		/* If mono, choose rising or falling CLK edge 
		 * based on the lr_pin_level */
		if(!config->lr_pin_level){
			/* If LOW, the dmic works as a LEFT channel only */
			pdm_param.gain_l = NRF_PDM_GAIN_MAXIMUM;
			pdm_param.edge = NRF_PDM_EDGE_LEFTFALLING;
		}else{
			/* If HIGH, the dmic works as a RIGHT channel only */
			pdm_param.gain_r = NRF_PDM_GAIN_MAXIMUM;
			pdm_param.edge = NRF_PDM_EDGE_LEFTRISING;
		}
	}

	audio_sample_rate = cfg->streams->pcm_rate;

	switch (audio_sample_rate)
	{
		case HZ_4800:
			dmic_data->inmp621_current_mode = LOW_POWER;
		#if NRF_PDM_HAS_RATIO_CONFIG
			/* For 4.8kHz we need ratio of 80x.
			* Otherwise we will put the sensor in STANDBY mode.
			* Resulting clock frequency is about 384kHz */
			pdm_param.ratio = NRF_PDM_RATIO_80X;
		#endif
			pdm_param.clock_freq = 0x030DB000UL;
			break;

		case HZ_6000:
			dmic_data->inmp621_current_mode = LOW_POWER;
		#if NRF_PDM_HAS_RATIO_CONFIG
			/* Resulting clock frequency is about 384kHz */
			pdm_param.ratio = NRF_PDM_RATIO_64X;
		#endif
			pdm_param.clock_freq = 0x030DB000UL;
			break;

		case HZ_8000:
			dmic_data->inmp621_current_mode = LOW_POWER;
		#if NRF_PDM_HAS_RATIO_CONFIG
			/* Resulting clock frequency is about 512kHz */
			pdm_param.ratio = NRF_PDM_RATIO_64X;
		#endif
			pdm_param.clock_freq = 0x04104000UL;
			break;

		case HZ_9600:
			dmic_data->inmp621_current_mode = LOW_POWER;
		#if NRF_PDM_HAS_RATIO_CONFIG
			/* Resulting clock frequency is about 614.4kHz */
			pdm_param.ratio = NRF_PDM_RATIO_64X;
		#endif
			pdm_param.clock_freq = 0x04DE5000UL;
			break;

		case HZ_11025:
			dmic_data->inmp621_current_mode = LOW_POWER;
		#if NRF_PDM_HAS_RATIO_CONFIG
			/* Resulting clock frequency is about 705.6kHz */
			pdm_param.ratio = NRF_PDM_RATIO_64X;
		#endif
			pdm_param.clock_freq = 0x05954000UL;
			break;

		case HZ_12000:
			dmic_data->inmp621_current_mode = LOW_POWER;
		#if NRF_PDM_HAS_RATIO_CONFIG
			/* Resulting clock frequency is about 768kHz */
			pdm_param.ratio = NRF_PDM_RATIO_64X;
		#endif
			pdm_param.clock_freq = 0x06123000UL;
			break;

		case HZ_16000:
			dmic_data->inmp621_current_mode = NORMAL;
		#if NRF_PDM_HAS_RATIO_CONFIG
			/* 1.28 MHz sampling with ratio 80x
			* This is equal to 16kHz output sample rate */
			pdm_param.ratio = NRF_PDM_RATIO_80X;
		#endif
			pdm_param.clock_freq = NRF_PDM_FREQ_1280K;
			break;

		case HZ_32000:
			dmic_data->inmp621_current_mode = NORMAL;
		#if NRF_PDM_HAS_RATIO_CONFIG
			/* Resulting clock frequency is about 2.048MHz */
			pdm_param.ratio = NRF_PDM_RATIO_64X;
		#endif
			pdm_param.clock_freq = 0x0FE03000UL;
			break;

		case HZ_44100:
			dmic_data->inmp621_current_mode = NORMAL;
		#if NRF_PDM_HAS_RATIO_CONFIG
			/* Resulting clock frequency is about 2.822MHz */
			pdm_param.ratio = NRF_PDM_RATIO_64X;
		#endif
			pdm_param.clock_freq = 0x15A02000UL;
			break;

		case HZ_48000:
			dmic_data->inmp621_current_mode = NORMAL;
		#if NRF_PDM_HAS_RATIO_CONFIG
			/* Resulting clock frequency is about 3.072MHz */
			pdm_param.ratio = NRF_PDM_RATIO_64X;
		#endif
			pdm_param.clock_freq = 0x17734000UL;
			break;

		default:
			break;
	}

	// nrfx_pdm_uninit();

	ret = nrfx_pdm_init(&pdm_param, inmp621_pdm_handler);
	if (ret != NRFX_SUCCESS) {
		LOG_ERR("Failed to inititialise the PDM peripheral, error %d", ret);
		return ret;
	}

	if(!irq_is_enabled(PDM0_IRQn)){
		irq_enable(PDM0_IRQn);
	}

	/* Copy the pdm struct in internal inmp621 data struct */
	memcpy(&dmic_data->inmp621_pdm_configs, &pdm_param, sizeof(pdm_param));

	return 0;
}

static int inmp621_get_command(const struct device *dev,
			       enum dmic_trigger cmd)
{
	int ret = 0;
	struct inmp621_data *dmic_data = dev->data;
	nrfx_pdm_config_t *pdm_param = &dmic_data->inmp621_pdm_configs;
	
	switch (cmd) {

		case DMIC_TRIGGER_STOP:
			/* Disable the pdm sampling and enter ZPL */
			ret = nrfx_pdm_stop();
			if (ret != NRFX_SUCCESS) {
				LOG_ERR("Failed to stop PDM sampling, error %d", ret);
                return ret;
			}
			dmic_data->inmp621_current_mode = ZPL;
			break;

		case DMIC_TRIGGER_PAUSE:
			/* Enter Standby mode */
			nrfx_pdm_uninit();

		#if NRF_PDM_HAS_RATIO_CONFIG
			pdm_param->ratio = NRF_PDM_RATIO_64X;
		#endif
			/* About 220kHz. Clock frequencies between
			 * 200kHz - 250kHz place the device in Standby. */
			pdm_param->clock_freq = 0x01C10000UL;

			ret = nrfx_pdm_init(pdm_param, inmp621_pdm_handler);
			if (ret != NRFX_SUCCESS) {
				LOG_ERR("Failed to inititialise the PDM peripheral, error %d", ret);
                return ret;
			}
			
			ret = nrfx_pdm_start();
			if (ret != NRFX_SUCCESS) {
				LOG_ERR("Failed to start PDM sampling, error %d", ret);
                return ret;
			}
			
			dmic_data->inmp621_current_mode = STANDBY;
			break;

		case DMIC_TRIGGER_START:
			/* The frequency is already set from inmp621_configure.
			 * If inmp621_configure was not first called, do nothing. */
			if(dmic_data->inmp621_current_mode == NORMAL || 
			   dmic_data->inmp621_current_mode == LOW_POWER){

				ret = nrfx_pdm_start();
				if (ret != NRFX_SUCCESS) {
					LOG_ERR("Failed to start PDM sampling, error %d", ret);
               		return ret;
				}
			}else{
				LOG_WRN("Must configure the inmp621 before starting sampling");
			}
			break;

		case DMIC_TRIGGER_RESET:
			//disable clock
			//clear buffer
			//clear dout
			ret = inmp621_clear_dout(dev);
			if (ret < 0) {
				LOG_ERR("Couldn't clear dout, error %d", ret);
           		return ret;

			}
			break;	

		default:
			break;
	}
	return 0;
}

static int inmp621_read_pdm(const struct device *dev, uint8_t stream,
			    void **buffer,
			    size_t *size, int32_t timeout)
{
	ARG_UNUSED(dev);
	ARG_UNUSED(stream);
	ARG_UNUSED(timeout);
	if(pcm_buffer_avalable) {
		/* Return pointer to 0th address of sampled buffer */
		*buffer = &pcm_buffer[pcm_buffer_in_use^1][0];
		/* Return the size of the buffer in bytes */
		*size = sizeof(pcm_buffer[0]);
		pcm_buffer_avalable = false;
	}
	else {
		*buffer = NULL;
		*size = 0;
	}
	return 0;
}
static const struct _dmic_ops inmp621_api_funcs = {
	.configure = inmp621_configure,
	.trigger = inmp621_get_command,
	.read  = inmp621_read_pdm,
};

static int inmp621_chip_init(const struct device *dev)
{
	uint8_t buff;
	int ret;

	/* Wait for ZPL to settle properly before
	 * initiating communication. */
	k_sleep(K_MSEC(1000));

	/* As the device does not contain a chip ID
	 * register, read the I2C_CNTRL register.
	 * The value 0x00 should return. */

	if(buff == 0x00){
		LOG_DBG("INMP621 dmic detected");
	}else{
		LOG_ERR("Device not found");
		return -ENXIO; 
	}
	
	ret = inmp621_wdt_configure(dev, INMP621_WDT_DELAY, IS_ENABLED(CONFIG_INMP621_WDT_ENABLE));
	if(ret){
		return ret;
	}

	ret = inmp621_bpf_set(dev, INMP621_LPF_FREQUENCY, INMP621_HPF_FREQUENCY);
	if(ret){
		return ret;
	}

	ret = inmp621_fast_mode_cnt_set(dev, INMP621_FAST_MODE_CNT);
	if(ret){
		return ret;
	}

	ret = inmp621_wos_rms_set(dev, CONFIG_INMP621_WOS_RMS_LPF);
	if(ret){
		return ret;
	}

	/* 45dB SPL is the average perceived loudness of a quiet library */
	ret = inmp621_wos_pga_min_thr_set(dev, (float)45);
	if(ret){
		return ret;
	}

	/* 90dB SPL is the average perceived loudness of a truck or loud speaker */
	ret = inmp621_wos_pga_max_thr_set(dev, (float)90);
	if(ret){
		return ret;
	}

	ret = inmp621_wos_dout_thresh_set(dev, CONFIG_INMP621_WOS_DOUT_THRESHOLD);
	if(ret){
		return ret;
	}

	return 0;
}

static int inmp621_init(const struct device *dev)
{
	const struct inmp621_config *config = dev->config;
	struct inmp621_data *dmic_data = dev->data;

	if (inmp621_chip_init(dev) < 0) {
		return -EINVAL;
	}

	dmic_data->inmp621_current_mode = ZPL;
	IRQ_CONNECT(PDM0_IRQn, 6, nrfx_pdm_irq_handler, NULL, 0);

	return 0;
}

DEVICE_DT_INST_DEFINE(0, inmp621_init, NULL,
		    &inmp621_data, &inmp621_cfg, POST_KERNEL,
		    CONFIG_AUDIO_DMIC_INIT_PRIORITY, &inmp621_api_funcs);
